use std::{ffi::OsString, os::fd::{AsRawFd, OwnedFd}};
use crate::{native, uapi::{self, i915::DrmGemTileInfo}};
use uapi::i915::DrmGemHandle;
use std::path::Path;
use std::fs::File;

pub use uapi::i915::GemIoctlError;

#[derive(Debug)]
pub struct DrmDeviceNode {
    pub fd: OwnedFd,
    pub path: Option<OsString>,
}

/// Owned GEM handle
/// Scoped to the device node file descriptor
/// (`drm_file->object_idr` in drm_file.h in the kernel)
#[derive(Debug)]
pub struct GemHandle<'a> {
    pub handle: DrmGemHandle,
    pub node: &'a DrmDeviceNode,
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
#[repr(u32)]
pub enum DrmCacheType {
    None    = native::I915_CACHING_NONE,
    Cached  = native::I915_CACHING_CACHED,
    Display = native::I915_CACHING_DISPLAY,
}

impl From<u32> for DrmCacheType {
    fn from(val: u32) -> DrmCacheType {
        match val {
            native::I915_CACHING_NONE => DrmCacheType::None,
            native::I915_CACHING_CACHED => DrmCacheType::Cached,
            native::I915_CACHING_DISPLAY => DrmCacheType::Display,
            _ => panic!("Invalid cache type encountered"),
        }
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub struct GemTileInfo {
    pub tiling_mode: u32,
    pub swizzle_mode: u32,
    pub phys_swizzle_mode: u32,
}

impl From<DrmGemTileInfo> for GemTileInfo {
    fn from(info: DrmGemTileInfo) -> GemTileInfo {
        GemTileInfo {
            tiling_mode: info.tiling_mode,
            swizzle_mode: info.swizzle_mode,
            phys_swizzle_mode: info.phys_swizzle_mode,
        }
    }
}

/// An owned GEM handle that will be closed when dropped
impl GemHandle<'_> {
    pub fn new(node: &DrmDeviceNode, size: u64) -> Option<GemHandle> {
        uapi::i915::make_gem(node.fd.as_raw_fd(), size).map(|handle| GemHandle {
            handle: handle,
            node: &node,
        })
    }

    pub fn from_handle(node: &DrmDeviceNode, handle: DrmGemHandle) -> Result<GemHandle, GemIoctlError> {
        // Avoid invoking GemHandle::is_valid() here to prevent the drop() method from trying to drop an invalid handle
        let is_valid = uapi::i915::gem_is_valid(node.fd.as_raw_fd(), &handle);

        if is_valid.is_ok() && is_valid.unwrap() {
            let res = GemHandle {
                handle: handle,
                node: &node,
            };
            return Ok(res);
        } else {
            return Err(is_valid.err().unwrap_or(GemIoctlError::InvalidHandle));
        }
    }

    pub fn close(self) -> Result<(), i32> {
        return unsafe { uapi::i915::close_gem_ref(self.node.fd.as_raw_fd(), &self.handle) };
    }

    // Should not be supported on Gen12+ GPUs
    // However, the i915 uAPI actually supports it on Gen12+
    // See https://patchwork.freedesktop.org/patch/325343/
    pub fn get_tiling(&self) -> Result<GemTileInfo, GemIoctlError> {
        uapi::i915::gem_get_tiling(self.node.fd.as_raw_fd(), &self.handle).map(|info| GemTileInfo::from(info))
    }

    // Only on iGPUs
    pub fn get_caching(&self) -> Result<DrmCacheType, GemIoctlError> {
        uapi::i915::gem_get_caching(self.node.fd.as_raw_fd(), &self.handle).map(|info| DrmCacheType::try_from(info).unwrap())
    }

    pub fn has_tiling(&self) -> Result<bool, uapi::i915::GemIoctlError> {
        uapi::i915::gem_has_tiling(self.node.fd.as_raw_fd(), &self.handle)
    }

    pub fn is_valid(&self) -> Result<bool, GemIoctlError> {
        uapi::i915::gem_is_valid(self.node.fd.as_raw_fd(), &self.handle)
    }
}

impl Drop for GemHandle<'_> {
    fn drop(&mut self) {
        // Ignoring the close failing as an invalid Gem handle's drop is a no-op
        _ = unsafe { uapi::i915::close_gem_ref(self.node.fd.as_raw_fd(), &self.handle) };
    }
}

#[derive(Debug)]
pub struct Device {
    pub node: DrmDeviceNode,
    pub driver: uapi::DrmVersion,
}

#[derive(Debug)]
pub struct Engine {
    pub info: uapi::i915::EngineInfo,
}

pub fn find_device() -> Option<Device> {
    uapi::i915::find_node().and_then(|file| Device::from_path(file.path?))
}

pub fn find_all_devices() -> Vec<Device> {
    uapi::i915::find_all_nodes().into_iter().filter_map(|file| Device::from_path(file.path.unwrap())).collect()
}

impl Device {
    pub fn from_fd(fd: OwnedFd) -> Option<Device> {
        let driver = uapi::get_drm_version(fd.as_raw_fd())?;
        Some(Device {
            node: DrmDeviceNode {
                fd: fd,
                path: None,
            },
            driver: driver,
        })
    }
    pub fn from_path(path: OsString) -> Option<Device> {
        let file = File::open(Path::new(&path)).ok()?;
        let driver = uapi::get_drm_version(file.as_raw_fd())?;
        Some(Device {
            node: DrmDeviceNode {
                fd: OwnedFd::from(file),
                path: Some(path),
            },
            driver: driver,
        })
    }
    pub fn get_engines(&self) -> Option<Vec<Engine>> {
        uapi::i915::get_engines(self.node.fd.as_raw_fd()).map(|engines| {
            engines.into_iter().map(|engine_info| {
                Engine {
                    info: engine_info,
                }
            }).collect()
        })
    }
    pub fn get_param(&self, param: u32) -> Option<i32> {
        uapi::i915::get_param(self.node.fd.as_raw_fd(), param as i32)
    }
}